'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.set = set;
exports.unset = unset;
exports.increment = increment;
exports.add = add;
exports.addUnique = addUnique;
exports.remove = remove;
exports.revert = revert;

var _arrayValueMatches = require('./arrayValueMatches');

var _arrayValueMatches2 = _interopRequireDefault(_arrayValueMatches);

var _deepCopy = require('./deepCopy');

var _deepCopy2 = _interopRequireDefault(_deepCopy);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/**
 * Set keys to specific values, or apply raw Ops
 */
function set(obj, changes) {
  var result = {};
  var serverData = obj._serverData || {};
  for (var _key in serverData) {
    result[_key] = serverData[_key];
  }
  Object.defineProperty(result, '_serverData', {
    enumerable: false,
    writable: false,
    configurable: false,
    value: serverData
  });

  var opSet = obj._opSet || {};
  var existing = false;
  for (var _key2 in obj) {
    result[_key2] = obj._serverData && obj._serverData.hasOwnProperty(_key2) ? obj._serverData[_key2] : obj[_key2];
    if (_key2 === 'objectId') {
      existing = true;
    } else if (!obj._opSet && !obj._serverData) {
      opSet[_key2] = obj[_key2];
    }
  }

  for (var _key3 in changes) {
    var op = changes[_key3];
    var source = void 0;
    if (op && op.__op) {
      switch (op.__op) {
        case 'Delete':
          delete result[_key3];
          if (existing) {
            opSet[_key3] = op;
          } else {
            delete opSet[_key3];
          }
          break;
        case 'Increment':
          var amount = isNaN(op.amount) ? 1 : op.amount;
          result[_key3] = (result[_key3] || 0) + amount;
          if (existing) {
            opSet[_key3] = op;
          } else {
            opSet[_key3] = result[_key3];
          }
          break;
        case 'Add':
          var toAdd = (0, _deepCopy2.default)(Array.isArray(op.objects) ? op.objects : [op.objects]);
          result[_key3] = (result[_key3] || []).concat(toAdd);
          if (existing) {
            opSet[_key3] = op;
          } else {
            opSet[_key3] = result[_key3];
          }
          break;
        case 'Remove':
          var toRemove = Array.isArray(op.objects) ? op.objects : [op.objects];
          var removed = [];
          source = result[_key3] || [];
          for (var i = 0; i < source.length; i++) {
            var matched = false;
            for (var j = 0; j < toRemove.length; j++) {
              if (matched) {
                continue;
              }
              if ((0, _arrayValueMatches2.default)(source[i], toRemove[j])) {
                matched = true;
              }
            }
            if (!matched) {
              removed.push(source[i]);
            }
          }
          result[_key3] = removed;
          if (existing) {
            opSet[_key3] = op;
          } else {
            opSet[_key3] = result[_key3];
          }
          break;
        case 'AddUnique':
          var toAddUnique = Array.isArray(op.objects) ? op.objects : [op.objects];
          var uniqueAdds = [];
          source = result[_key3] || [];
          for (var _i = 0; _i < toAddUnique.length; _i++) {
            var _matched = false;
            for (var _j = 0; _j < source.length; _j++) {
              if (_matched) {
                continue;
              }
              if ((0, _arrayValueMatches2.default)(toAddUnique[_i], source[_j])) {
                _matched = true;
              }
            }
            if (!_matched) {
              uniqueAdds.push((0, _deepCopy2.default)(toAddUnique[_i]));
            }
          }
          var merged = source.concat(uniqueAdds);
          result[_key3] = merged;
          if (existing) {
            opSet[_key3] = op;
          } else {
            opSet[_key3] = result[_key3];
          }
          break;
        default:
          throw new Error('Unsupported Op');
      }
    } else {
      // Standard set op
      result[_key3] = (0, _deepCopy2.default)(op);
      opSet[_key3] = (0, _deepCopy2.default)(op);
    }
  }
  Object.defineProperty(result, '_opSet', {
    enumerable: false,
    writable: false,
    configurable: false,
    value: opSet
  });

  return result;
}

/**
 * Remove a field from an object
 */
function unset(obj, key) {
  return set(obj, _defineProperty({}, key, { __op: 'Delete' }));
}

/**
 * Atomically increment (or decrement) a numeric field
 */
function increment(obj, key, amount) {
  var inc = typeof amount === 'number' ? amount : 1;
  return set(obj, _defineProperty({}, key, { __op: 'Increment', amount: inc }));
}

/**
 * Atomically add elements to an Array field
 */
function add(obj, key, objects) {
  if (!Array.isArray(objects)) {
    objects = [objects];
  }
  return set(obj, _defineProperty({}, key, { __op: 'Add', objects: objects }));
}

/**
 * Atomically add elements to an Array field, if they don't already exist
 */
function addUnique(obj, key, objects) {
  if (!Array.isArray(objects)) {
    objects = [objects];
  }
  return set(obj, _defineProperty({}, key, { __op: 'AddUnique', objects: objects }));
}

/**
 * Atomically remove elements from an Array field
 */
function remove(obj, key, objects) {
  if (!Array.isArray(objects)) {
    objects = [objects];
  }
  return set(obj, _defineProperty({}, key, { __op: 'Remove', objects: objects }));
}

/**
 * Client-side operation to revert all unsaved changes
 */
function revert(obj, key) {
  if (!obj._serverData) {
    return {};
  }
  var original = {};
  Object.defineProperty(original, '_serverData', {
    enumerable: false,
    writable: false,
    configurable: false,
    value: obj._serverData
  });
  Object.defineProperty(original, '_opSet', {
    enumerable: false,
    writable: false,
    configurable: false,
    value: {}
  });
  for (var _key4 in obj._serverData) {
    original[_key4] = obj._serverData[_key4];
  }
  return original;
}