'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.emptyQuery = emptyQuery;
exports.equalTo = equalTo;
exports.notEqualTo = notEqualTo;
exports.lessThan = lessThan;
exports.lessThanOrEqualTo = lessThanOrEqualTo;
exports.greaterThan = greaterThan;
exports.greaterThanOrEqualTo = greaterThanOrEqualTo;
exports.containedIn = containedIn;
exports.notContainedIn = notContainedIn;
exports.containsAll = containsAll;
exports.exists = exists;
exports.doesNotExist = doesNotExist;
exports.matches = matches;
exports.matchesQuery = matchesQuery;
exports.doesNotMatchQuery = doesNotMatchQuery;
exports.matchesKeyInQuery = matchesKeyInQuery;
exports.doesNotMatchKeyInQuery = doesNotMatchKeyInQuery;
exports.contains = contains;
exports.startsWith = startsWith;
exports.endsWith = endsWith;
exports.near = near;
exports.withinRadians = withinRadians;
exports.withinMiles = withinMiles;
exports.withinKilometers = withinKilometers;
exports.withinGeoBox = withinGeoBox;
exports.ascending = ascending;
exports.descending = descending;
exports.skip = skip;
exports.limit = limit;
exports.include = include;
exports.select = select;
exports.or = or;
exports.find = find;
exports.get = get;

var _WireFormat = require('./WireFormat');

var _deepCopy = require('./deepCopy');

var _deepCopy2 = _interopRequireDefault(_deepCopy);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var CONSTRAINTS = {
  $ne: true,
  $lt: true,
  $lte: true,
  $gt: true,
  $gte: true,
  $in: true,
  $nin: true,
  $all: true,
  $exists: true,
  $regex: true,
  $options: true,
  $inQuery: true,
  $notInQuery: true,
  $select: true,
  $dontSelect: true,
  $nearSphere: true,
  $within: true
};

function isConstraint(value) {
  for (var c in CONSTRAINTS) {
    if (value.hasOwnProperty(c)) {
      return true;
    }
  }
  return false;
}

function quote(s) {
  return '\\Q' + s.replace('\\E', '\\E\\\\E\\Q') + '\\E';
}

function emptyQuery() {
  return {
    where: {},
    order: [],
    limit: -1,
    skip: 0,
    keys: null,
    include: []
  };
}

function copyQuery(q) {
  var copy = {
    where: q.where ? (0, _deepCopy2.default)(q.where) : {},
    order: q.order ? [].concat(q.order) : [],
    limit: q.hasOwnProperty('limit') ? q.limit : -1,
    skip: q.skip || 0,
    keys: q.keys ? [].concat(q.keys) : null,
    include: q.include ? [].concat(q.include) : []
  };
  return copy;
}

function copyWithNewConstraint(q, field, constraint, value) {
  var copy = copyQuery(q);
  var encoded = (0, _WireFormat.encode)(value, true);
  if (copy.where.hasOwnProperty(field)) {
    if (!isConstraint(copy.where[field])) {
      copy.where[field] = {};
    }
  } else {
    copy.where[field] = {};
  }
  copy.where[field][constraint] = encoded;
  return copy;
}

function equalTo(q, field, value) {
  if (typeof value === 'undefined') {
    return doesNotExist(q, field);
  }
  var copy = copyQuery(q);
  copy.where[field] = (0, _WireFormat.encode)(value);
  return copy;
}

function notEqualTo(q, field, value) {
  return copyWithNewConstraint(q, field, '$ne', value);
}

function lessThan(q, field, value) {
  return copyWithNewConstraint(q, field, '$lt', value);
}

function lessThanOrEqualTo(q, field, value) {
  return copyWithNewConstraint(q, field, '$lte', value);
}

function greaterThan(q, field, value) {
  return copyWithNewConstraint(q, field, '$gt', value);
}

function greaterThanOrEqualTo(q, field, value) {
  return copyWithNewConstraint(q, field, '$gte', value);
}

function containedIn(q, field, value) {
  return copyWithNewConstraint(q, field, '$in', value);
}

function notContainedIn(q, field, value) {
  return copyWithNewConstraint(q, field, '$nin', value);
}

function containsAll(q, field, value) {
  return copyWithNewConstraint(q, field, '$all', value);
}

function exists(q, field) {
  return copyWithNewConstraint(q, field, '$exists', true);
}

function doesNotExist(q, field) {
  return copyWithNewConstraint(q, field, '$exists', false);
}

function matches(q, field, regex) {
  var modifiers = arguments.length <= 3 || arguments[3] === undefined ? '' : arguments[3];

  var copy = copyWithNewConstraint(q, field, '$regex', regex);
  if (regex.ignoreCase) {
    modifiers += 'i';
  }
  if (regex.multiline) {
    modifiers += 'm';
  }
  if (modifiers.length) {
    copy.where[field].$options = modifiers;
  }
  return copy;
}

function matchesQuery(q, field, className, query) {
  return copyWithNewConstraint(q, field, '$inQuery', _extends({}, query, { className: className }));
}

function doesNotMatchQuery(q, field, className, query) {
  return copyWithNewConstraint(q, field, '$notInQuery', _extends({}, query, { className: className }));
}

function matchesKeyInQuery(q, field, className, query, queryKey) {
  return copyWithNewConstraint(q, field, '$select', {
    key: queryKey,
    query: _extends({}, query, { className: className })
  });
}

function doesNotMatchKeyInQuery(q, field, className, query, queryKey) {
  return copyWithNewConstraint(q, field, '$dontSelect', {
    key: queryKey,
    query: _extends({}, query, { className: className })
  });
}

function contains(q, field, value) {
  if (typeof value !== 'string') {
    throw new Error('The value being searched for must be a string.');
  }
  return copyWithNewConstraint(q, field, '$regex', quote(value));
}

function startsWith(q, field, value) {
  if (typeof value !== 'string') {
    throw new Error('The value being searched for must be a string.');
  }
  return copyWithNewConstraint(q, field, '$regex', '^' + quote(value));
}

function endsWith(q, field, value) {
  if (typeof value !== 'string') {
    throw new Error('The value being searched for must be a string.');
  }
  return copyWithNewConstraint(q, field, '$regex', quote(value) + '$');
}

function near(q, field, point) {
  return copyWithNewConstraint(q, field, '$nearSphere', point);
}

function withinRadians(q, field, point, distance) {
  var copy = copyWithNewConstraint(q, field, '$nearSphere', point);
  copy.where[field].$maxDistance = distance;
  return copy;
}

function withinMiles(q, field, point, distance) {
  return withinRadians(q, field, point, distance / 3958.8);
}

function withinKilometers(q, field, point, distance) {
  return withinRadians(q, field, point, distance / 6371.0);
}

function withinGeoBox(q, field, southwest, northeast) {
  return copyWithNewConstraint(q, field, '$within', {
    $box: [southwest, northeast]
  });
}

function ascending(q) {
  var copy = copyQuery(q);

  for (var _len = arguments.length, keys = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    keys[_key - 1] = arguments[_key];
  }

  copy.order = keys.map(function (k) {
    if (typeof k !== 'string') {
      throw new Error('Ascending should only take strings as secondary arguments');
    }
    return k;
  });
  return copy;
}

function descending(q) {
  var copy = copyQuery(q);

  for (var _len2 = arguments.length, keys = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
    keys[_key2 - 1] = arguments[_key2];
  }

  copy.order = keys.map(function (k) {
    if (typeof k !== 'string') {
      throw new Error('Descending should only take strings as secondary arguments');
    }
    return '-' + k;
  });
  return copy;
}

function skip(q, n) {
  if (typeof n !== 'number' || n < 0) {
    throw new Error('You can only skip by a positive number');
  }
  var copy = copyQuery(q);
  copy.skip = n;
  return copy;
}

function limit(q, n) {
  if (typeof n !== 'number' || n < 0) {
    throw new Error('You can only skip by a positive number');
  }
  var copy = copyQuery(q);
  copy.limit = n;
  return copy;
}

function include(q) {
  var copy = copyQuery(q);

  for (var _len3 = arguments.length, keys = Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
    keys[_key3 - 1] = arguments[_key3];
  }

  copy.include = keys;
  return copy;
}

function select(q) {
  throw new Error('not yet implemented');
}

function or() {
  for (var _len4 = arguments.length, queries = Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
    queries[_key4] = arguments[_key4];
  }

  if (queries.length === 0) {
    throw new Error('At least one input query is required');
  }

  var q = emptyQuery();
  q.where = {
    $or: [queries.map(function (q) {
      return (0, _deepCopy2.default)(q.where);
    })]
  };
  return q;
}

function find(app, className, q) {
  var options = arguments.length <= 3 || arguments[3] === undefined ? {} : arguments[3];

  var filtered = {
    where: q.where
  };
  if (q.order && q.order.length) {
    filtered.order = q.order.join(',');
  }
  if (q.limit > -1) {
    filtered.limit = q.limit;
  }
  if (q.skip > 0) {
    filtered.skip = q.skip;
  }
  if (q.keys && q.keys.length) {
    filtered.keys = q.keys.join(',');
  }
  if (q.include && q.include.length) {
    filtered.include = q.include.join(',');
  }
  return app.client.get('classes/' + className, filtered, options).then(function (response) {
    var results = response.response.results;
    var objects = results.map(function (r) {
      if (!r) {
        return r;
      }
      var obj = (0, _WireFormat.decode)(r);
      Object.defineProperty(obj, '_serverData', {
        enumerable: false,
        writeable: false,
        configurable: false,
        value: (0, _WireFormat.decode)(r)
      });
      Object.defineProperty(obj, '_className', {
        enumerable: false,
        writeable: false,
        configurable: false,
        value: className
      });
      return obj;
    });
    return objects;
  });
}

function get(app, className, objectId) {
  var options = arguments.length <= 3 || arguments[3] === undefined ? {} : arguments[3];

  return app.client.get('classes/' + className + '/' + objectId, {}, options).then(function (response) {
    var result = response.response;
    if (!result) {
      return result;
    }
    var obj = (0, _WireFormat.decode)(result);
    Object.defineProperty(obj, '_serverData', {
      enumerable: false,
      writeable: false,
      configurable: false,
      value: (0, _WireFormat.decode)(result)
    });
    Object.defineProperty(obj, '_className', {
      enumerable: false,
      writeable: false,
      configurable: false,
      value: className
    });
    return obj;
  });
};